/*  pack: an install wrapper that logs all files installed for easy            
	  uninstalling.
    Copyright (C) 1998 Matt Perry <guy@somewhere.linuxos.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "pack.h"

struct Option {
	char *name;
	int  arg_type;
	int  ret_val;
};

char *bad_opt;

static Option pack_options[] = {
	{"new",				2, 'N'},
	{"list",			2, 'L'},
	{"done",			2, 'D'},
	{"erase",			1, 'E'},

	{"help",			0, 'h'},
	{"quiet",			0, 'q'},
	{"force",			0, 'f'},
	{"version",			0, 'v'},
	{"no-checksum",			0, '5'},

	{0, 0, 0}
};

static Option install_options[] = {
	{"directory",			0, 'd'},
	{"group",			1, 'g'},
	{"mode",			1, 'm'},
	{"owner",			1, 'o'},
	{"version-control",		1, 'V'},
	{"suffix",			1, 'S'},
		
	{0, 0, 0}
};

void short_match(int &, char** &, int &);
void long_match(int &, char** &, int &);
void handle_opt(Option*, int, char*);
void fix_args(char** &, int &, char*);

void process_args(int &argc, char** &argv)
{
	bad_opt = 0;
	argind = 1;

	int i = 1;
	while (argv[i]) {
		/* -x */
		if (argv[i][0] == '-' && argv[i][1] != '-') {
			short_match(argc, argv, i);
		}
		/* --xxxxx */
		else if (argv[i][0] == '-' && argv[i][1] == '-') {
			if (argv[i][2] == 0) { /* plain -- */
				/* argv[0]: dummy value */
				fix_args(argv, i, argv[0]);
				break;
			}

			long_match(argc, argv, i);
		}
		else // only i++ here, since the _match's do it for us.
			++i;
	}

	if (packmode && bad_opt) {
		fprintf(stderr, "%s: invalid option: '%s'\n",
			progname, bad_opt);
		exit(1);
	}
}

void short_match(int &argc, char** &argv, int &argi)
{
	Option* list = pack_options;
	char* opt = argv[argi]+1;
	char* optarg = 0;
	int i = 0;
	bool found = false;

	for (;;) {
		if (list[i].name == 0) {
			if (list == install_options)
				break;
			list = install_options;
			i = 0;
			continue;
		}

		if (list[i].ret_val != opt[0]) {
			++i;
			continue;
		}

		if (list[i].arg_type == 1) {
			if (opt[1])
				optarg = opt+1;
			else if (argv[argi+1])
				optarg = argv[argi+1];
			else {
				fprintf(stderr, "%s: '-%c' requires argument.\n",
					progname, opt[0]);
				exit(1);
			}
			handle_opt(list, i, optarg);
		}
		else if (list[i].arg_type == 2) {
			if (opt[1])
				optarg = opt+1;
			handle_opt(list, i, optarg);
		}
		else
			handle_opt(list, i, NULL);

		found = true;
		break;
	}

	/* We don't want to whine just yet.  First make sure it's intended
	 * for Pack and not install. */
	if (!found && !bad_opt)
		bad_opt = argv[argi];

	if (found && list == pack_options) {
		mremove(argc, argv, argi);
	} else
		fix_args(argv, argi, optarg);
}

void long_match(int &argc, char** &argv, int &argi)
{
	Option* list = pack_options;
	int i = 0;
	bool found = false;

	// This part splits the option and its argument.
	char *opt = (char*)strdup(argv[argi]+2);
	char *optarg;
	if ((optarg = strchr(opt, '=')) != NULL)
		*optarg++ = 0;

	for (;;) {
		if (list[i].name == 0) {
			if (list == install_options)
				break;
			list = install_options;
			i = 0;
			continue;
		}

		if (!strmat(list[i].name, opt)) {
			++i;
			continue;
		}

		if (list[i].arg_type == 1) {
			if (!optarg && argv[argi+1])
				optarg = argv[argi+1];
			if (!optarg) {
				fprintf(stderr, "%s: '--%s' requires argument.\n",
					progname, opt);
				exit(1);
			}
			handle_opt(list, i, optarg);
		}
		else if (list[i].arg_type == 2)
			handle_opt(list, i, optarg);
		else
			handle_opt(list, i, NULL);

		found = true;
		break;
	}

	/* We don't want to whine just yet.  First make sure it's intended
	 * for Pack and not install. */
	if (!found && !bad_opt)
		bad_opt = argv[argi];
	else
		free(opt);

	if (found && list == pack_options) {
		mremove(argc, argv, argi);
	} else
		fix_args(argv, argi, optarg);
}

void handle_opt(Option* list, int i, char* optarg)
{
	if (optarg && optarg[0] == 0)
		optarg = 0;

	if (list[i].ret_val == 'd')
		p_obj->flags |= FLAG_DIR_ARG;

	if (list == install_options)
		return;		// No one cares.

	switch (list[i].ret_val) {
		case 'N':
		case 'L':
		case 'D':
		case 'E':
			if (packmode)
				needhelp();
			p_obj->cmdarg = (optarg ? (char*)strdup(optarg) : 0);
			packmode = list[i].ret_val;
			break;
		case 'h':
			usage();
			break;
		case 'q':
			p_obj->flags &= ~FLAG_VERBOSE;
			break;
		case 'f':
			p_obj->flags &= ~FLAG_PROMPT_ERASE;
			break;
		case 'v':
			version();
			break;
		case '5':
			p_obj->flags &= ~FLAG_CHKSUM;
			break;
		default:
			fprintf(stderr, "Error 01: val %d\n", i);
			needhelp();
			break;
	}
}

void fix_args(char** &argv, int &argi, char* optarg)
{
	char* hold;
	for (;;) {
		if (argi != argind) {
			hold = argv[argi];
			for (int j = argi; j > argind; --j)
				argv[j] = argv[j-1];
			argv[argind] = hold;
		}
		++argind, ++argi;

		/* if nextarg == optarg, rearrange that one too. */
		if (optarg != argv[argi])
			break;
	}
}
