/*  pack: an install wrapper that logs all files installed for easy
	  uninstalling.
    Copyright (C) 1998 Matt Perry <guy@somewhere.linuxos.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <unistd.h>
#include <dirent.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>

#define _mainfile
#include "pack.h"

int main(int argc, char *argv[])
{
	char* p;
	PackObj obj(argc, argv);

	progname = strdup(argv[0]);
	p_obj = &obj;

	if (argc < 2)
		needhelp();

	install_path = PACK_INSTALL_PATH;
	if ((p=getenv("PACK_INSTALL_PATH")) != NULL)
		install_path = (char*)strdup(p);

	chksum_path = PACK_CHKSUM_PATH;
	if ((p=getenv("PACK_CHKSUM_PATH")) != NULL)
		chksum_path = (char*)strdup(p);

	db_dir = PACK_DB_DIR;
	if ((p=getenv("PACK_DB_DIR")) != NULL)
		db_dir = (char*)strdup(p);

	packmode = 0;
	process_args(argc, argv);

	if (chksum_path[0] == 0)
		obj.flags &= ~(FLAG_CHKSUM);

	switch (packmode) {
	case 'N':
		obj.cmd_new();
		break;
	case 'L':
		obj.cmd_list();
		break;
	case 'D':
		obj.cmd_done();
		break;
	case 'E':
		obj.cmd_erase();
		break;
	case 0:	/* Install */
		if (argc-argind < 1)
			needhelp();
		obj.cmd_install();
		break;
	}

	return 0;
}

// class PackObj {
PackObj::PackObj(int argc, char** argv) : argc(argc), argv(argv),
	flags(FLAG_VERBOSE | FLAG_PROMPT_ERASE | FLAG_CHKSUM)
{
}

void PackObj::cmd_new()
{
	FILE* fp;
	if ((fp = fopen(Path(db_dir, ".open"), "a")) == NULL) {
		fprintf(stderr, "%s: can't write to '%s': ", progname, gpath);
		perror("");	exit(1);
	}

	fprintf(stderr,
		"NOTE: Make sure you are in the toplevel dir of the source tree that you\n"
		"are installing from before runnning '%s -N'\n\n", argv[0]);

	char* curdir = getcwd(NULL, 4096);
	if (cmdarg)
		packcur = cmdarg;
	else
		packcur = Path(curdir).lastdir();

	if ((get_packcur(packcur)) == 0) {
		fprintf(fp, "%s\t%s\n", curdir, packcur);
		fclose(fp);
	}

	struct stat sbuf;
	if ((stat(Path(db_dir, packcur), &sbuf)) < 0) {
		printf("Creating new package: '%s'\n", packcur);

		if ((fp = fopen(Path(db_dir, packcur), "w")) == NULL) {
			fprintf(stderr, "%s: can't write to '%s': ", progname, gpath);
			perror("");	exit(1);
		}
		fclose(fp);
	}
	else
		printf("Appending to existing package: '%s'\n", packcur);
}

void PackObj::cmd_list()
{
	if (cmdarg == 0) {
		DIR* dirp;
		struct dirent* entry;

		if ((dirp = opendir(db_dir)) == NULL) {
			fprintf(stderr, "%s: can't read dir '%s': ", progname, db_dir);
			perror("");	exit(1);
		}

		while ((entry = readdir(dirp)) != NULL) {
			if (entry->d_name[0] == '.')
				continue;
			
			if ((get_packcur(entry->d_name)) != 0)
				printf("%s - OPEN\n", entry->d_name);
			else
				printf("%s\n", entry->d_name);
		}
		closedir(dirp);
		return;
	}

	FILE* fp;
	if ((fp = fopen(Path(db_dir, cmdarg), "r")) == NULL) {
		fprintf(stderr, "%s: can't read '%s': ", progname, gpath);
		perror("");	exit(1);
	}

	char buf[4096], info[4096]="", path[4096]="";
	while ((fgets(buf, sizeof(buf), fp)) != NULL) {
		/* info: if md5sum was enabled, info is either checksum info (for
		 * files) or 'd' (for dirs).
		 * If md5sum was not enabled, info is either 'f' (for files) or
		 * 'd' (for dirs). */
		sscanf(buf, "%[^ ]  %s", info, path);
		if (info[0] == 'd' && info[1] == 0)
			printf("-d %s\n", path);
		else if (!(flags & FLAG_CHKSUM))
			printf("%s\n", path);
	}
	fclose(fp);

	if ((flags & FLAG_CHKSUM)) {
		sprintf(buf, "%s -c %s 2>&1", chksum_path, (char*)Path(db_dir, cmdarg));
		system(buf);
	}
	else
		printf("WARNING: checksum information is disabled: checksums not verified.\n");
}

void PackObj::cmd_done()
{
	char buf1[4096], buf2[4096];

	if (rename(Path(db_dir, ".open", buf1), Path(db_dir, ".open~", buf2)) < 0) {
		fprintf(stderr, "%s: can't access package logfile\n", progname);
		perror("");	exit(1);
	}

	FILE *fp, *wp;
	if ((fp = fopen(Path(db_dir, ".open~"), "r")) == NULL) {
		fprintf(stderr, "%s: can't open package log~\n", progname);
		perror("");	exit(1);
	}
	if ((wp = fopen(Path(db_dir, ".open"), "w")) == NULL) {
		fprintf(stderr, "%s: can't open package log\n", progname);
		perror("");	exit(1);
	}

	char *packfind;
	if (cmdarg)
		packfind = cmdarg;
	else {
		packfind = getcwd(NULL, 4096);
		packfind = Path(packfind).lastdir();
	}

	char dirstr[4096], packstr[512];
	bool gotone=false, wroteone=false;
	while (1) {
		if ((fscanf(fp, "%s\t%s\n", dirstr, packstr)) < 2) {
			if (!gotone) {
				if (!cmdarg)
					fprintf(stderr, "%s: no open package from this directory.\n", progname);
				else
					fprintf(stderr, "%s: wasn't working with package '%s'\n", progname, cmdarg);
				rename(Path(db_dir, ".open~", buf1),
				       Path(db_dir, ".open", buf2));
				exit(1);
			}
			else
				break;
		}

		if (strmat(packstr, packfind))
			gotone = true;
		else {
			fprintf(wp, "%s\t%s\n", dirstr, packstr);
			wroteone = true;
		}
	}

	fclose(fp);
	fclose(wp);

	if ((unlink(Path(db_dir, ".open~"))) < 0) {
		fprintf(stderr, "%s: can't remove '%s': ", progname, gpath);
		perror("");
	}
	if (!wroteone && (unlink(Path(db_dir, ".open"))) < 0) {
		fprintf(stderr, "%s: can't remove '%s': ", progname, gpath);
		perror("");
	}
}

void PackObj::cmd_erase()
{
	FILE* fp;
	if ((fp = fopen(Path(db_dir, cmdarg), "r")) == NULL) {
		fprintf(stderr, "%s: can't read '%s': ", progname, gpath);
		perror("");	exit(1);
	}

	// it's a while loop so that "L" will come back to the prompt.
	char buf[4096];
	while ((flags & FLAG_PROMPT_ERASE)) {
		printf("Uninstall package '%s': Y/N/L/V (yes)/(no)/(list contents)/(verify checksum): ", cmdarg);
		fgets(buf, sizeof(buf), stdin);
		if (toupper(buf[0]) == 'Y')
			break;
		else if (toupper(buf[0]) == 'N')
			return;
		else if (toupper(buf[0]) == 'L')
			this->cmd_list();
	}

	verify_chksum();

	packcur = get_packcur(cmdarg);
	if (packcur)
		cmd_done();

	char info[4096]="", path[4096]="";
	struct stat sbuf;
	bool failed=false;
	Buffer<char*> dirs;

	while ((fgets(buf, sizeof(buf), fp)) != NULL) {
		if ((sscanf(buf, "%[^ ]  %s", info, path)) < 2) {
			fprintf(stderr, "%s: bad packlog entry '%s'\n", progname, buf);
			continue;
		}

		if ((stat(path, &sbuf)) < 0) {
			V( printf("file missing - %s\n", path); );
			continue;
		}

		if (info[0] == 'd' && info[1] == 0) {
			dirs.addlen(+1);
			dirs.data[dirs.len] = (char*)strdup(Path(path, ""));
		}
		else {
			if (unlink(path) < 0) {
				printf("del FAILED - %s: %s\n", path, strerror(errno));
				failed = true;
				continue;
			}

			V( printf("del - %s\n", path); );

			/* Add this guy's dir to the list of dirs to remove. */
			strcpy(gpath, path); Path(gpath).filename()[0] = 0;
			stat(gpath, &sbuf);

			struct stat sbuf2;
			bool found = false;
			for (int i=0; dirs.data[i]; ++i) {
				stat(dirs.data[i], &sbuf2);
				if (sbuf.st_ino == sbuf2.st_ino) {
					found = true;
					break;
				}
			}

			/* If we didn't find a match, add one. */
			if (!found) {
				dirs.addlen(+1);
				dirs.data[dirs.len] = (char*)strdup(gpath);
			}
		}
	}

	dirs.data[++dirs.len] = 0;
	for (int i = 0; dirs.data[i]; ++i) {
		if (rmdir(dirs.data[i]) >= 0) {
			V( printf("deldir - %s\n", dirs.data[i]); );
		} else if (errno == ENOTEMPTY) {
			printf("deldir FAILED - %s: not empty\n", dirs.data[i]);
		} else {
			printf("deldir FAILED - %s: %s\n", dirs.data[i], strerror(errno));
		}
	}

	/* Don't delete the package file - one or more files weren't removed. */
	if (failed)
		return;

	if (unlink(Path(db_dir, cmdarg)) < 0) {
		fprintf(stderr, "%s: can't remove package file '%s': ", progname, cmdarg);
		perror("");	exit(1);
	}
}

void PackObj::cmd_install()
{
	char** src, **dst, **p;
	int i;

	packcur = get_packcur();
	if (packcur == 0) {
		fprintf(stderr, "%s: NO OPEN PACKAGE (create one with pack -N)\n", progname);
		exit(1);
	}

	src = mnew(char*, argc);
	dst = mnew(char*, argc);

	/* if '-d', all the args are dests */
	if ((flags & FLAG_DIR_ARG)) {
		src[0] = 0;
		for (p = dst, i = argind; argv[i]; ++i, ++p) {
			p[0] = argv[i];
			p[1] = 0;
		}

		install_dirs(src, dst);
	}
	/* otherwise, only the last one is dest, the ones before are sources. */
	else {
		for (p = src, i = argind; argv[i+1]; ++i, ++p) {
			p[0] = argv[i];
			p[1] = 0;
		}
		dst[0] = argv[i];		dst[1] = 0;

		install_files(src, dst);
	}

	/* This executes PACK_INSTALL_PATH with specified arguments. It also
	 * changes around argv and argind, but since argind is global and
	 * argv is in this class, we see those changes (we want that) */
	execute();

	/* Write the checksums of the files to a log.  This is used during
	 * deletion, so we don't erase modified files. */
	write_chksum();
}

void PackObj::install_dirs(char** src, char** dst)
{
	struct stat sbuf;
	int i;

	for (i = 0; dst[i]; ++i) {
		/* Only add to database if the dir doesn't exist! This
		 * has caused problems in the past from packages running
		 * "install -d /usr/bin". */
		if ((stat(dst[i], &sbuf)) < 0)
			add_entry(dst[i]);
	}
}

void PackObj::install_files(char **src, char **dst)
{
	struct stat sbuf;
	char path[4096];
	char* p, *ppath;
	int i;

	if (dst[0] == 0) {
		fprintf(stderr, "%s: destination not specified\n", progname);
		exit(1);
	}

	stat(dst[0], &sbuf);
	if (!S_ISDIR(sbuf.st_mode)) {
		add_entry(dst[0]);
		return;
	}

	ppath = Path(dst[0], "", path);
	ppath += strlen(ppath);
	for (i = 0; src[i]; ++i) {
		stat(src[i], &sbuf);
		if (S_ISDIR(sbuf.st_mode))
			continue;

		p = Path(src[i]).filename();
		strcpy(ppath, p);

		add_entry(path);
	}
}

void PackObj::add_entry(char* path)
{
	FILE* fp;

	V( printf("add - %s\n", path); );

	/* If it's a dir, always put "d <path>" in the packlog.  If it's a file,
	 * check either put "f <path>" or "<checksum> <path>" depending on
	 * whether or not md5sum is enabled. */
	if ((flags & FLAG_DIR_ARG) || !(flags & FLAG_CHKSUM)) {
		if ((fp = fopen(Path(db_dir, packcur), "a")) == NULL) {
			fprintf(stderr, "%s: can't write to '%s': ", progname, gpath);
			perror("");		exit(1);
		}

		fprintf(fp, "%s  %s\n", ((flags & FLAG_DIR_ARG) ? "d" : "f"), path);
		fclose(fp);
	}
	else {
		if (chksum.len < 0)
			chksum.len = 0;

		int curpos = chksum.len;
		chksum.addlen(strlen(path)+1);

		strcpy(&chksum.data[curpos], path);
		chksum.data[chksum.len-1] = ' ';
		chksum.data[chksum.len  ] = 0;
	}
}

void PackObj::write_chksum()
{
	if (chksum.data[0] == 0 || !(flags & FLAG_CHKSUM))
		return;

	/* Stupid ANSI C++ doesn't allow variable length arrays... */
	char* buf = new char[strlen(chksum.data)+strlen(chksum_path)+32];

	sprintf(buf, "%s %s >> %s", chksum_path, chksum.data, (char*)Path(db_dir, packcur));
	system(buf);

	delete buf;
}

void PackObj::verify_chksum()
{
	if (!(flags & FLAG_CHKSUM))
		return;

	FILE *fp;
	char buf[4096];

	sprintf(buf, "%s -c %s 2>&1", chksum_path, (char*)Path(db_dir, cmdarg));

	if ((fp = popen(buf, "r")) == NULL) {
		fprintf(stderr, "%s: Couldn't run checksum verifier '%s -c': ", progname, chksum_path);
		perror("");	exit(1);
	}

	char path[4096], check[4096];
	bool failed=false;
	while ((fgets(buf, sizeof(buf), fp)) != NULL) {
		/* Don't see why scanf should fail, but we'll check just in case. */
		if ((sscanf(buf, "%[^:]: %[^: \n\t]", path, check)) < 2) {
			fprintf(stderr, "%s: md5sum output is unrecognized.\n", progname);
			continue;
		}
		/* This is just md5sum's warnings: we don't care. */
		if (strmat(path, chksum_path) &&
			!(strmat(check, "FAILED") || strmat(check, "OK")))
			continue;

		if (!strmat(check, "OK")) {
			printf("checksum %s - %s\n", path, check);
			failed = true;
		}
	}
	pclose(fp);

	if (failed) {
		printf("Files have been changed since installation. Uninstall cancelled.\n");
		printf("To uninstall anyway, use 'install --no-checksum -E<package>'\n");
		exit(1);
	}
}

// packfind(=NULL): the package name you're looking for.
char* PackObj::get_packcur(char* packfind)
{
	FILE* fp;
	if ((fp = fopen(Path(db_dir, ".open"), "r")) == NULL)
		return 0;

	char* curdir;
	if (!packfind)
		curdir = getcwd(NULL, 4096);

	char dirstr[4096];
	static char packstr[512];
	while (1) {
		if ((fscanf(fp, "%s\t%s\n", dirstr, packstr)) < 2)
			return 0;

		/* This is kinda wierd: if packfind is specified, we're
		 * looking for a certain package name and we don't care
		 * what dir we're in.  Otherwise, we only want to return
		 * a package if we're in the right directory. */
		if (packfind && strmat(packfind, packstr))
			break;
		// found one that works for curdir
		else if (!packfind && strpre(curdir, dirstr))
			break;
	}
	if (!packfind)
		free(curdir);
	fclose(fp);

	return packstr;
}

int PackObj::execute()
{
	/* Remove pack arguments */
	argv[0] = install_path;

	int pid = fork();
	switch (pid) {
	case -1:
		fprintf(stderr, "%s: fork() failed!\n", progname);
		break;
	case 0: {		/* Child. */
		execv(argv[0], argv);
		exit(127);
		break;
	}
	default:		/* Parent. */
		int status;
		while (pid != wait(&status))	/* Wait for kid to finish. */
			;
		return status;
	}
	return -1;
}
// class PackObj }

// class Path {
Path::Path(char* dir, char* file, char* buf) : pathstr(buf)
{
	while ((*buf = *dir))
		++buf, ++dir;

	if (buf[-1] != '/')
		*buf++ = '/';

	while ((*buf = *file))
		++buf, ++file;
}

Path::Path(char* path) : pathstr(path)
{ }

char* Path::filename()
{
	char* p;

	if ((p = strrchr(pathstr, '/')) != NULL)
		++p;
	else
		p = pathstr;

	return p;
}

/* Get the last directory in the path string.
 */
char* Path::lastdir()
{
	char* p = &pathstr[strlen(pathstr)-1];

	while (*p == '/' || *p == '.')
		--p;

	while (p > pathstr && *p != '/')
		--p;
	if (*p == '/')
		++p;

	return p;
}

#if 0
// NOT USED
/* Fix any references to './' and '../' in pathstr.
 */
char* Path::demangle()
{
	char buf[4096];
	char *p=pathstr, *b=buf;

	while (*p) {
		if (p[0] == '.' && p[1] == '.' && p[2] == '/' && p[-1] == '/') {
			p += 3;
			--b;
			while (*--b != '/')
				;
			*++b = 0;
		}
		else if (p[0] == '.' && p[1] == '/' && p[-1] == '/')
			p += 2;
		else {
			*b++ = *p++;
			*b = 0;
		}
	}

	if (*--b == '/')
		*b = 0;

	strcpy(pathstr, buf);
	return pathstr;
}
#endif
// class Path }

// class Buffer {
template <class T>
Buffer<T>::Buffer() : len(-1), size(MEM_BLOCK)
{
	data = mnew(T, size);
}

template <class T>
void Buffer<T>::addlen(int alen)
{
	len += alen;
	if (len >= size-2) {
		size += MEM_BLOCK;
		data = mrenew(T, data, size);
	}
}
// class Buffer }

int strpre(char* hay, char* needle)
{
	for (; *needle; ++hay, ++needle) {
		if (*hay != *needle)
			return 0;
	}
	return 1;
}

void version(void)
{
	fprintf(stderr, "Pack install monitor (Matt Perry) %d.%d.%d\n",
		PACK_VER_MAJOR, PACK_VER_MINOR, PACK_VER_PATCH);
	exit(1);
}

void needhelp(void)
{
	fprintf(stderr, "Try '%s --help' for more information.\n", progname);
	exit(1);
}

void usage(void)
{
	static char usage[] =
"Usage: %s [options] <command>\n"
"options:\n"
"  -h, --help\n"
"     Print this help message.\n"
"  -q, --quiet\n"
"     Turn quiet mode on. Changes to the package are not printed.\n"
"  -5, --no-checksum\n"
"     Don't call md5sum to verify or log checksums.\n"
"  -f, --force\n"
"     Uninstall without prompting (--erase command).\n"
"\n"
"Commands:\n"
"  -N, --new[=PACKAGE]:\n"
"     Open a new package.  You must be in the toplevel directory of the\n"
"     installation's source tree when issuing this command.  If PACKAGE\n"
"     is specified, call the new package PACKAGE.  Otherwise, the package\n"
"     takes on the name of the current directory.\n"
"  -L, --list[=PACKAGE]:\n"
"     If PACKAGE is specified, list all files in that package. Otherwise\n"
"     list all packages.\n"
"  -D, --done[=PACKAGE]:\n"
"     If PACKAGE is specified, close that package.  Otherwise, close the\n"
"     current package (this is determined by the current directory).\n"
"  -E, --erase=PACKAGE:\n"
"     Remove all files listed in package PACKAGE, and delete PACKAGE.\n"
"\nOR...\n"
"Usage: %s [options] INSTALL_ARGUMENTS\n"
"     '%s'\n"
"     is executed with INSTALL_ARGUMENTS passed to it. All files\n"
"     installed are saved in the current package (create one with -N).\n";

	fprintf(stderr, usage, progname, progname, install_path);
	exit(1);
}
